@@ -4,8 +4,12 @@
 # This source code is licensed under the BSD-style license found in the
 # LICENSE file in the root directory of this source tree.
 
-import datetime
 import json
+import shlex
+import shutil
+import time
+
+import jwt
 import os.path
 import subprocess
 from http import HTTPStatus
@@ -15,6 +19,8 @@ from grader_service.autograding.grader_executor import GraderExecutor
 from grader_service.autograding.local_feedback import GenerateFeedbackExecutor
 from grader_service.handlers.handler_utils import parse_ids
 import tornado
+from tornado.httpclient import AsyncHTTPClient, HTTPRequest, HTTPClientError
+from tornado.escape import url_escape, json_decode
 from grader_service.api.models.submission import Submission as SubmissionModel
 from grader_service.orm.assignment import AutoGradingBehaviour
 from grader_service.orm.submission import Submission
@@ -23,15 +29,19 @@ from grader_service.registry import VersionSpecifier, register_handler
 from sqlalchemy.sql.expression import func
 from tornado.web import HTTPError
 from grader_convert.gradebook.models import GradeBookModel
+from subprocess import PIPE, CalledProcessError
+from tornado.process import Subprocess
+
 
 from grader_service.handlers.base_handler import GraderBaseHandler, authorize, RequestHandlerConfig
+import datetime
 
 
 def tuple_to_submission(t, student=False):
     """
     Transforms tuple with values into a submission entity.
 
-    :param student: does the tuple contain submissions for the student
+    :param student: if True, get submissions for student where some data is not needed
     :param t: tuple with values
     :return: submission entity
     """
@@ -45,6 +55,7 @@ def tuple_to_submission(t, student=False):
         s.assignid,
         s.commit_hash,
         s.feedback_available,
+        s.edited,
         s.logs,
         s.date,
     ) = t
@@ -56,6 +67,7 @@ def tuple_to_submission(t, student=False):
 
     return s
 
+
 @register_handler(
     path=r"\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/submissions\/?",
     version_specifier=VersionSpecifier.ALL,
@@ -114,12 +126,13 @@ class SubmissionHandler(GraderBaseHandler):
                         Submission.assignid,
                         Submission.commit_hash,
                         Submission.feedback_available,
+                        Submission.edited,
                         Submission.logs,
                         func.max(Submission.date),
                     )
-                        .filter(Submission.assignid == assignment_id)
-                        .group_by(Submission.username)
-                        .all()
+                    .filter(Submission.assignid == assignment_id)
+                    .group_by(Submission.username)
+                    .all()
                 )
                 submissions = [tuple_to_submission(t) for t in submissions]
             elif submission_filter == 'best':
@@ -133,12 +146,13 @@ class SubmissionHandler(GraderBaseHandler):
                         Submission.assignid,
                         Submission.commit_hash,
                         Submission.feedback_available,
+                        Submission.edited,
                         Submission.logs,
                         Submission.date,
                     )
-                        .filter(Submission.assignid == assignment_id)
-                        .group_by(Submission.username)
-                        .all()
+                    .filter(Submission.assignid == assignment_id, Submission.feedback_available == True)
+                    .group_by(Submission.username)
+                    .all()
                 )
                 submissions = [tuple_to_submission(t) for t in submissions]
             else:
@@ -155,15 +169,16 @@ class SubmissionHandler(GraderBaseHandler):
                         Submission.assignid,
                         Submission.commit_hash,
                         Submission.feedback_available,
+                        Submission.edited,
                         Submission.logs,
                         func.max(Submission.date),
                     )
-                        .filter(
+                    .filter(
                         Submission.assignid == assignment_id,
                         Submission.username == role.username,
                     )
-                        .group_by(Submission.username)
-                        .all()
+                    .group_by(Submission.username)
+                    .all()
                 )
                 submissions = [tuple_to_submission(t, True) for t in submissions]
             elif submission_filter == 'best':
@@ -177,15 +192,16 @@ class SubmissionHandler(GraderBaseHandler):
                         Submission.assignid,
                         Submission.commit_hash,
                         Submission.feedback_available,
+                        Submission.edited,
                         Submission.logs,
                         Submission.date,
                     )
-                        .filter(
+                    .filter(
                         Submission.assignid == assignment_id,
                         Submission.username == role.username,
                     )
-                        .group_by(Submission.username)
-                        .all()
+                    .group_by(Submission.username)
+                    .all()
                 )
                 submissions = [tuple_to_submission(t, True) for t in submissions]
             else:
@@ -193,11 +209,11 @@ class SubmissionHandler(GraderBaseHandler):
                     self.session.query(
                         Submission
                     )
-                        .filter(
+                    .filter(
                         Submission.assignid == assignment_id,
                         Submission.username == role.username,
                     )
-                        .all()
+                    .all()
                 )
         if response_format == "csv":
             # csv format does not include logs
@@ -412,9 +428,10 @@ class SubmissionPropertiesHandler(GraderBaseHandler):
         try:
             gradebook = GradeBookModel.from_dict(json.loads(properties_string))
 
-            score = gradebook.score + self.get_extra_credit(gradebook)
+            score = gradebook.score
 
-        except:
+        except Exception as e:
+            self.log.info(e)
             raise HTTPError(HTTPStatus.BAD_REQUEST, reason="Cannot parse properties file!")
         submission.score = score
         submission.properties = properties_string
@@ -426,6 +443,224 @@ class SubmissionPropertiesHandler(GraderBaseHandler):
         for notebook in gradebook.notebooks.values():
             for grades in notebook.grades:
                 extra_credit += grades.extra_credit if grades.extra_credit is not None else 0
-        # TODO change score to float
         self.log.info("Extra credit is " + str(extra_credit))
-        return int(extra_credit)
+        return extra_credit
+
+
+@register_handler(
+    path=r"\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/submissions\/(?P<submission_id>\d*)\/edit\/?",
+    version_specifier=VersionSpecifier.ALL,
+)
+class SubmissionEditHandler(GraderBaseHandler):
+        
+    @authorize([Scope.tutor, Scope.instructor])
+    async def put(self, lecture_id: int, assignment_id: int, submission_id: int):
+        '''
+        Create or overwrites the repository which stores changes of submissions files
+        :param lecture_id: lecture id
+        :param assignment_id: assignment id
+        :param submission_id: submission id
+        :return:
+        '''
+        lecture_id, assignment_id = parse_ids(lecture_id, assignment_id)
+        self.validate_parameters()
+
+        submission = self.get_submission(lecture_id, assignment_id, submission_id)
+        assignment = submission.assignment
+        lecture = assignment.lecture
+
+        # Path to repository which will store edited submission files
+        git_repo_path = os.path.join(
+            self.gitbase,
+            lecture.code,
+            str(assignment.id),
+            "edit",
+            str(submission_id),
+        )
+
+        # Path to repository of student which contains the submitted files
+        submission_repo_path = os.path.join(
+            self.gitbase,
+            lecture.code,
+            str(assignment.id),
+            assignment.type,
+            submission.username
+        )
+
+        if os.path.exists(git_repo_path):
+            shutil.rmtree(git_repo_path)
+
+        # Creating bare repository
+        if not os.path.exists(git_repo_path):
+            os.makedirs(git_repo_path, exist_ok=True)
+        
+        self._run_command(f'git init --bare', git_repo_path)
+
+        # Create temporary paths to copy the submission files in the edit repository
+        tmp_path = os.path.join(
+            self.application.grader_service_dir,
+            "tmp",
+            lecture.code,
+            str(assignment.id),
+            "edit",
+            str(submission.id),
+        )
+    
+
+        tmp_input_path = os.path.join(
+            tmp_path,
+            "input"
+        )
+
+        tmp_output_path = os.path.join(
+            tmp_path,
+            "output"
+        )
+
+        if os.path.exists(tmp_path):
+            shutil.rmtree(tmp_path, ignore_errors=True)
+
+        os.makedirs(tmp_input_path, exist_ok=True)
+
+        # Init local repository
+        command = f"git init"
+        self._run_command(command, tmp_input_path)
+
+        # Pull user repository
+        command = f'git pull "{submission_repo_path}" main'
+        self._run_command(command, tmp_input_path)
+        self.log.info("Successfully cloned repo")
+
+        # Checkout to correct submission commit
+        command = f"git checkout {submission.commit_hash}"
+        self._run_command(command, tmp_input_path)
+        self.log.info(f"Now at commit {submission.commit_hash}")
+
+        # Copy files to output directory
+        shutil.copytree(tmp_input_path,tmp_output_path, ignore = shutil.ignore_patterns(".git"))
+
+        # Init local repository
+        command = f"git init"
+        self._run_command(command, tmp_output_path)
+
+        # Add edit remote
+        command = f"git remote add edit {git_repo_path}"
+        self._run_command(command, tmp_output_path)
+        self.log.info("Successfully added edit remote")
+
+         # Switch to main
+        command = f"git switch -c main"
+        self._run_command(command, tmp_output_path)
+        self.log.info("Successfully switched to branch main")
+        
+
+        # Add files to staging
+        command = f"git add -A"
+        self._run_command(command, tmp_output_path)
+        self.log.info("Successfully added files to staging")
+
+        # Commit Files
+        command = f'git commit -m "Initial commit" '
+        self._run_command(command, tmp_output_path)
+        self.log.info("Successfully commited files")
+
+        # Push copied files
+        command = f"git push edit main"
+        self._run_command(command, tmp_output_path)
+        self.log.info("Successfully pushed copied files")
+
+        submission.edited = True
+        self.session.commit()
+        self.write_json(submission)
+        
+
+
+@register_handler(
+    path=r"\/lectures\/(?P<lecture_id>\d*)\/assignments\/(?P<assignment_id>\d*)\/submissions\/lti\/?",
+    version_specifier=VersionSpecifier.ALL,
+)
+class LtiSyncHandler(GraderBaseHandler):
+    cache_token = {"token": None, "ttl": datetime.datetime.now()}
+
+    @authorize([Scope.instructor])
+    async def get(self, lecture_id: int, assignment_id: int):
+        submissions = (
+            self.session.query(
+                Submission.id,
+                Submission.username,
+                Submission.score,
+                func.max(Submission.date)
+            )
+            .filter(Submission.assignid == assignment_id, Submission.auto_status == "automatically_graded",
+                    Submission.feedback_available == True)
+            .group_by(Submission.username)
+            .all()
+        )
+
+        assignment = self.get_assignment(lecture_id, assignment_id)
+
+        lti_username_convert = RequestHandlerConfig.instance().lti_username_convert
+
+        if lti_username_convert is None:
+            raise HTTPError(HTTPStatus.NOT_FOUND, reason="Unable to match users: lti_username_convert is not set in grader "
+                                                  "config")
+
+        scores = [{"id": s[0], "username": lti_username_convert(s[1]), "score": s[2]} for s in submissions]
+        stamp = datetime.datetime.now()
+        if LtiSyncHandler.cache_token["token"] and LtiSyncHandler.cache_token["ttl"] > stamp - datetime.timedelta(
+                minutes=50):
+            token = LtiSyncHandler.cache_token["token"]
+        else:
+            token = await self.request_bearer_token()
+            LtiSyncHandler.cache_token["token"] = token
+            LtiSyncHandler.cache_token["ttl"] = datetime.datetime.now()
+
+        scores = {"assignment": assignment, "scores": scores, "token": token}
+
+        self.write_json(scores)
+
+    async def request_bearer_token(self):
+        # get config variables
+        lti_client_id = RequestHandlerConfig.instance().lti_client_id
+        if lti_client_id is None:
+            raise HTTPError(HTTPStatus.NOT_FOUND,
+                            reason="Unable to request bearer token: lti_client_id is not set in grader config")
+        lti_token_url = RequestHandlerConfig.instance().lti_token_url
+        if lti_token_url is None:
+            raise HTTPError(HTTPStatus.NOT_FOUND,
+                            reason="Unable to request bearer token: lti_token_url is not set in grader config")
+
+        private_key = RequestHandlerConfig.instance().lti_token_private_key
+        if private_key is None:
+            raise HTTPError(HTTPStatus.NOT_FOUND,
+                            reason="Unable to request bearer token: lti_token_private_key is not set in grader config")
+        if callable(private_key):
+            private_key = private_key()
+
+        payload = {"iss": "grader-service", "sub": lti_client_id, "aud": [lti_token_url],
+                   "ist": str(int(time.time())), "exp": str(int(time.time()) + 60),
+                   "jti": str(int(time.time())) + "123"}
+        try:
+            encoded = jwt.encode(payload, private_key, algorithm="RS256")
+        except Exception as e:
+            raise HTTPError(HTTPStatus.UNPROCESSABLE_ENTITY, f"Unable to encode payload: {str(e)}" )
+        self.log.info("encoded: " + encoded)
+        scopes = [
+            "https://purl.imsglobal.org/spec/lti-ags/scope/score",
+            "https://purl.imsglobal.org/spec/lti-ags/scope/lineitem",
+            "https://purl.imsglobal.org/spec/lti-nrps/scope/contextmembership.readonly"
+        ]
+        scopes = url_escape(" ".join(scopes))
+        data = f"grant_type=client_credentials&client_assertion_type=urn%3Aietf%3Aparams%3Aoauth%3Aclient-assertion" \
+               f"-type%3Ajwt-bearer&client_assertion={encoded}&scope={scopes} "
+        self.log.info("data: " + data)
+
+        httpclient = AsyncHTTPClient()
+        try:
+            response = await httpclient.fetch(HTTPRequest(url=lti_token_url, method="POST", body=data,
+                                                          headers={
+                                                              "Content-Type": "application/x-www-form-urlencoded"}))
+        except HTTPClientError as e:
+            self.log.error(e.response)
+            raise HTTPError(e.code, reason="Unable to request token:" + e.response.reason)
+        return json_decode(response.body)["access_token"]
